/**
 * vim: tabstop=2:shiftwidth=2:softtabstop=2:expandtab
 * Plugin demonstrating the use of
 * https://community.openvpn.net/openvpn/ticket/15
 *
 * It simulates a user session where
 *
 * authentication occurs
 * user bandwidth consumption is updated to accounting server regularly
 * session close
 *
 * compile with 
 * gcc -shared -Wl,-soname,accounting-test.so -o \
 * accounting-test.so testplugin.c
 */

/*
 * server conf:
 * plugin /path/to/accounting-test.so
 *
 * server need to be configured with --enable-accounting
 * which is available in patch attached to 
 * https://community.openvpn.net/openvpn/ticket/15
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <time.h>


#include "openvpn-plugin.h"

/*
 * Our context, where we keep our state.
 */

struct plugin_context{
  char *context;
};

char *strdupf (const char *fmt, ...);
const char *get_env (const char *name, const char *envp[]);

OPENVPN_EXPORT openvpn_plugin_handle_t
openvpn_plugin_open_v2 (unsigned int *type_mask, const char *argv[],
                        const char *envp[], struct openvpn_plugin_string_list **return_list)
{
  struct plugin_context *context;
	int rc = 0;


  /*
   * Allocate our context
   */
  context = malloc (sizeof (struct plugin_context));

  *type_mask = OPENVPN_PLUGIN_MASK ( OPENVPN_PLUGIN_CLIENT_CONNECT_V2) | 
                OPENVPN_PLUGIN_MASK ( OPENVPN_PLUGIN_CLIENT_DISCONNECT ) |
                OPENVPN_PLUGIN_MASK ( OPENVPN_PLUGIN_AUTH_USER_PASS_VERIFY ) |
                OPENVPN_PLUGIN_MASK ( OPENVPN_PLUGIN_ACCOUNTING );
  return (openvpn_plugin_handle_t) context;
}

OPENVPN_EXPORT int
openvpn_plugin_func_v2 (openvpn_plugin_handle_t handle, const int type,
                        const char *argv[], const char *envp[],
                        void *per_client_context,  struct openvpn_plugin_string_list **return_list)
{
  struct plugin_context *context = (struct plugin_context *) handle;
  int acc_time;
  time_t now;
  char *now_str;

  time(&now);
  now_str = ctime(&now);
  if (type == OPENVPN_PLUGIN_CLIENT_CONNECT_V2){
    acc_time = rand()%20 + 1;
    *return_list = malloc (sizeof (struct openvpn_plugin_string_list));
    (*return_list)->next = NULL;
    (*return_list)->name = strdup ("config");
    (*return_list)->value = strdupf("accounting-freq %d", acc_time);
    printf ("ACCT %s Sending start time to accounting server %s\n", now_str, get_env("time_unix", envp));
    printf ("ACCT %s Plugin requested accounting update in %d seconds\n", now_str, acc_time);
    return OPENVPN_PLUGIN_FUNC_SUCCESS;
  }
  if (type == OPENVPN_PLUGIN_ACCOUNTING){
      const char *tx, *rx;
      tx = get_env("bytes_sent", envp);
      rx = get_env("bytes_received", envp);
      printf ("ACCT %s Sending accounting information to server %stx/%srx\n", now_str, tx ? tx : "0", rx ? rx : "0");
      *return_list = malloc (sizeof (struct openvpn_plugin_string_list));
      (*return_list)->next = NULL;
      (*return_list)->name = strdup ("config");
      acc_time = rand() % 20 + 1;
      (*return_list)->value = strdupf("accounting-freq %d", acc_time);
      printf("ACCT %s Acounting server requesting new update in %d seconds\n", now_str, acc_time);

    return OPENVPN_PLUGIN_FUNC_SUCCESS;
  }

  if (type == OPENVPN_PLUGIN_CLIENT_DISCONNECT){
      const char *tx, *rx;
      tx = get_env("bytes_sent", envp);
      rx = get_env("bytes_received", envp);

    printf ("ACCT %s Sending accounting information to server %stx/%srx, session started at %s and last %s\n", 
              now_str, tx ? tx : "0", rx ? rx : "0",
              get_env("time_unix", envp), get_env("time_duration", envp));
    return OPENVPN_PLUGIN_FUNC_SUCCESS;
  }
  if (type == OPENVPN_PLUGIN_AUTH_USER_PASS_VERIFY){
    return OPENVPN_PLUGIN_FUNC_SUCCESS;
  } 
	return OPENVPN_PLUGIN_FUNC_ERROR;
}

/**
 * plugin_close_v1:
 * Called immediately prior to plug-in unload
 */
OPENVPN_EXPORT void
openvpn_plugin_close_v1 (openvpn_plugin_handle_t handle)
{
  struct plugin_context *context = (struct plugin_context *) handle;
  free (context);
}

const char *
get_env (const char *name, const char *envp[])
{
  if (envp){
    int i;
    const int namelen = strlen (name);
    for (i = 0; envp[i]; ++i){
      if (!strncmp (envp[i], name, namelen))
	    {
	      const char *cp = envp[i] + namelen;
	      if (*cp == '=')
		      return cp + 1;
	    }
	  }
  }
  return NULL;
}



char *
strdupf (const char *fmt, ...){
  va_list  vargs;
  char     buf[BUFSIZ];
  char    *p;

  if (!fmt) {
    return (NULL);
  }
  va_start (vargs, fmt);
  vsnprintf (buf, sizeof (buf), fmt, vargs);
  va_end (vargs);

  buf[sizeof (buf) - 1] = '\0';        /* ensure buf is NUL-terminated */

  if (!(p = strdup (buf))) {
    return (NULL);
  }
  return (p);
}


